from math import ceil
from typing import Any, cast

from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.decorators.cache import never_cache
from django.views.decorators.http import require_POST
from rq import requeue_job
from rq.exceptions import NoSuchJobError
from rq.job import Job, JobStatus
from rq.registry import (
    DeferredJobRegistry,
    FailedJobRegistry,
    FinishedJobRegistry,
    ScheduledJobRegistry,
    StartedJobRegistry,
)
from rq.worker import Worker
from rq.worker_registration import clean_worker_registry

from .queues import get_queue_by_index, get_scheduler_by_index
from .settings import get_queues_map
from .utils import get_executions, get_jobs, stop_jobs


def rq_viewname(request: HttpRequest, viewname: str) -> str:
    current_app = getattr(request, "current_app", None)
    if current_app:
        return f"{current_app}:django_rq_{viewname}"
    return f"django_rq:{viewname}"


def each_context(request: HttpRequest) -> dict[str, Any]:
    return {
        **admin.site.each_context(request),
    }


@never_cache
@staff_member_required
def jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    items_per_page = 100
    num_jobs = queue.count
    page = int(request.GET.get('page', 1))

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        jobs = queue.get_jobs(offset, items_per_page)
    else:
        jobs = []
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Queued',
    }
    return render(request, 'django_rq/jobs.html', context_data)


@never_cache
@staff_member_required
def finished_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    registry = FinishedJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))

    if request.GET.get('desc', '1') == '1':
        sort_direction = 'descending'
    else:
        sort_direction = 'ascending'

    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, offset + items_per_page - 1, desc=sort_direction == 'descending')
        jobs = get_jobs(queue, job_ids, registry)

    else:
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'sort_direction': sort_direction,
    }
    return render(request, 'django_rq/finished_jobs.html', context_data)


@never_cache
@staff_member_required
def failed_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    registry = FailedJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))

    if request.GET.get('desc', '1') == '1':
        sort_direction = 'descending'
    else:
        sort_direction = 'ascending'

    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, offset + items_per_page - 1, desc=sort_direction == 'descending')
        jobs = get_jobs(queue, job_ids, registry)

    else:
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'sort_direction': sort_direction,
    }
    return render(request, 'django_rq/failed_jobs.html', context_data)


@never_cache
@staff_member_required
def scheduled_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    registry = ScheduledJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []

    if request.GET.get('desc', '1') == '1':
        sort_direction = 'descending'
    else:
        sort_direction = 'ascending'

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, offset + items_per_page - 1, desc=sort_direction == 'descending')
        jobs = get_jobs(queue, job_ids, registry)
        for job in jobs:
            job.scheduled_at = registry.get_scheduled_time(job)  # type: ignore[attr-defined]
    else:
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'sort_direction': sort_direction,
    }
    return render(request, 'django_rq/scheduled_jobs.html', context_data)


@never_cache
@staff_member_required
def started_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    registry = StartedJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []
    executions = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)

        try:
            composite_keys = registry.get_job_and_execution_ids(offset, offset + items_per_page - 1)
        except AttributeError:
            composite_keys = [
                cast(tuple[str, str], key.split(':'))
                for key in registry.get_job_ids(offset, offset + items_per_page - 1)
            ]

        jobs = get_jobs(queue, [i[0] for i in composite_keys], registry)
        executions = get_executions(queue, composite_keys)

    else:
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Started',
        'executions': executions,
    }
    return render(request, 'django_rq/started_job_registry.html', context_data)


@never_cache
@staff_member_required
def workers(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    clean_worker_registry(queue)
    all_workers = Worker.all(queue.connection)
    workers = [worker for worker in all_workers if queue.name in worker.queue_names()]

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'workers': workers,
    }
    return render(request, 'django_rq/workers.html', context_data)


@never_cache
@staff_member_required
def worker_details(request: HttpRequest, queue_index: int, key: str) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    worker = Worker.find_by_key(key, connection=queue.connection)
    assert worker
    # Convert microseconds to milliseconds
    worker.total_working_time = worker.total_working_time / 1000

    queue_names = ', '.join(worker.queue_names())

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'worker': worker,
        'queue_names': queue_names,
        'job': worker.get_current_job(),
        'total_working_time': worker.total_working_time * 1000,
    }
    return render(request, 'django_rq/worker_details.html', context_data)


@never_cache
@staff_member_required
def deferred_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    registry = DeferredJobRegistry(queue.name, queue.connection)

    items_per_page = 100
    num_jobs = len(registry)
    page = int(request.GET.get('page', 1))
    jobs = []

    if request.GET.get('desc', '1') == '1':
        sort_direction = 'descending'
    else:
        sort_direction = 'ascending'

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        job_ids = registry.get_job_ids(offset, offset + items_per_page - 1, desc=sort_direction == 'descending')
        for job_id in job_ids:
            try:
                jobs.append(Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer))
            except NoSuchJobError:
                pass

    else:
        page_range = []

    context_data = {
        **each_context(request),
        'queue': queue,
        'queue_index': queue_index,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
        'job_status': 'Deferred',
        'sort_direction': sort_direction,
    }
    return render(request, 'django_rq/deferred_jobs.html', context_data)


@never_cache
@staff_member_required
def job_detail(request: HttpRequest, queue_index: int, job_id: str) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    try:
        job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)
    except NoSuchJobError:
        raise Http404(f"Couldn't find job with this ID: {job_id}")

    try:
        job.func_name
        data_is_valid = True
    except:
        data_is_valid = False

    # Backward compatibility support for RQ < 1.12.0
    rv = job.connection.hget(job.key, 'result')
    if rv is not None:
        # cache the result
        job.legacy_result = job.serializer.loads(rv)  # type: ignore[attr-defined]
    try:
        exc_info = job._exc_info
    except AttributeError:
        exc_info = None

    dependencies = []
    # if job._dependency_ids:
    # Fetch dependencies if they exist
    # dependencies = Job.fetch_many(
    #     job._dependency_ids, connection=queue.connection, serializer=queue.serializer
    # )
    for dependency_id in job._dependency_ids:
        try:
            dependency = Job.fetch(dependency_id, connection=queue.connection, serializer=queue.serializer)
        except NoSuchJobError:
            dependency = None
        dependencies.append((dependency_id, dependency))

    dependents = []
    for dependent_id in job.dependent_ids:
        try:
            dependent = Job.fetch(dependent_id, connection=queue.connection, serializer=queue.serializer)
        except NoSuchJobError:
            dependent = None
        dependents.append((dependent_id, dependent))

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
        'data_is_valid': data_is_valid,
        'exc_info': exc_info,
        'dependencies': dependencies,
        'dependents': dependents,
    }
    return render(request, 'django_rq/job_detail.html', context_data)


@never_cache
@staff_member_required
def result_detail(request: HttpRequest, queue_index: int, job_id: str, result_id: str) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    try:
        job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)
    except NoSuchJobError:
        raise Http404(f"Couldn't find job with this ID: {job_id}")

    result = next((result for result in job.results() if result.id == result_id), None)
    if not result:
        raise Http404(f"Couldn't find result with this ID: {result_id}")

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
        'result': result,
    }
    return render(request, 'django_rq/result_detail.html', context_data)


@never_cache
@staff_member_required
def delete_job(request: HttpRequest, queue_index: int, job_id: str) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)

    if request.method == 'POST':
        # Remove job id from queue and delete the actual job
        queue.connection.lrem(queue.key, 0, job.id)
        job.delete()
        messages.info(request, f'You have successfully deleted {job.id}')
        return redirect(rq_viewname(request, "jobs"), queue_index)

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/delete_job.html', context_data)


@never_cache
@staff_member_required
def requeue_job_view(request: HttpRequest, queue_index: int, job_id: str) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)

    if request.method == 'POST':
        requeue_job(job_id, connection=queue.connection, serializer=queue.serializer)
        messages.info(request, f'You have successfully requeued {job.id}')
        return redirect(rq_viewname(request, "job_detail"), queue_index, job_id)

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/delete_job.html', context_data)


@never_cache
@staff_member_required
def clear_queue(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)

    if request.method == 'POST':
        queue.empty()
        messages.info(request, f'You have successfully cleared the queue {queue.name}')
        return redirect(rq_viewname(request, "jobs"), queue_index)

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'queue': queue,
    }
    return render(request, 'django_rq/clear_queue.html', context_data)


@never_cache
@staff_member_required
def requeue_all(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    registry = FailedJobRegistry(queue=queue)

    if request.method == 'POST':
        job_ids = registry.get_job_ids()
        count = 0
        # Confirmation received
        for job_id in job_ids:
            try:
                requeue_job(job_id, connection=queue.connection, serializer=queue.serializer)
                count += 1
            except NoSuchJobError:
                pass

        messages.info(request, 'You have successfully requeued %d jobs!' % count)
        return redirect(rq_viewname(request, "jobs"), queue_index)

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'queue': queue,
        'total_jobs': len(registry),
    }

    return render(request, 'django_rq/requeue_all.html', context_data)


@never_cache
@staff_member_required
def delete_failed_jobs(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    registry = FailedJobRegistry(queue=queue)

    if request.method == 'POST':
        job_ids = registry.get_job_ids()
        jobs = Job.fetch_many(job_ids, connection=queue.connection)
        count = 0
        for job in jobs:
            if job:
                job.delete()
                count += 1

        messages.info(request, 'You have successfully deleted %d jobs!' % count)
        return redirect(rq_viewname(request, "home"))

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'queue': queue,
        'total_jobs': len(registry),
    }

    return render(request, 'django_rq/clear_failed_queue.html', context_data)


@never_cache
@staff_member_required
def confirm_action(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    next_url = request.META.get('HTTP_REFERER') or reverse(rq_viewname(request, "jobs"), args=[queue_index])

    if request.method == 'POST' and request.POST.get('action', False):
        # confirm action
        if request.POST.get('_selected_action', False):
            context_data = {
                **each_context(request),
                'queue_index': queue_index,
                'action': request.POST['action'],
                'job_ids': request.POST.getlist('_selected_action'),
                'queue': queue,
                'next_url': next_url,
            }
            return render(request, 'django_rq/confirm_action.html', context_data)

    return redirect(next_url)


@never_cache
@staff_member_required
def actions(request: HttpRequest, queue_index: int) -> HttpResponse:
    queue = get_queue_by_index(queue_index)
    next_url = request.POST.get('next_url') or reverse(rq_viewname(request, "jobs"), args=[queue_index])

    if request.method == 'POST' and request.POST.get('action', False):
        # do confirmed action
        if request.POST.get('job_ids', False):
            job_ids = request.POST.getlist('job_ids')

            if request.POST['action'] == 'delete':
                for job_id in job_ids:
                    job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)
                    # Remove job id from queue and delete the actual job
                    queue.connection.lrem(queue.key, 0, job.id)
                    job.delete()
                messages.info(request, f'You have successfully deleted {len(job_ids)} jobs!')
            elif request.POST['action'] == 'requeue':
                for job_id in job_ids:
                    requeue_job(job_id, connection=queue.connection, serializer=queue.serializer)
                messages.info(request, 'You have successfully requeued %d  jobs!' % len(job_ids))
            elif request.POST['action'] == 'stop':
                stopped, failed_to_stop = stop_jobs(queue, job_ids)
                if len(stopped) > 0:
                    messages.info(request, 'You have successfully stopped %d jobs!' % len(stopped))
                if len(failed_to_stop) > 0:
                    messages.error(request, '%d jobs failed to stop!' % len(failed_to_stop))

    return redirect(next_url)


@never_cache
@staff_member_required
def enqueue_job(request: HttpRequest, queue_index: int, job_id: str) -> HttpResponse:
    """Enqueue deferred jobs"""
    queue = get_queue_by_index(queue_index)
    job = Job.fetch(job_id, connection=queue.connection, serializer=queue.serializer)

    if request.method == 'POST':
        try:
            # _enqueue_job is new in RQ 1.14, this is used to enqueue
            # job regardless of its dependencies
            queue._enqueue_job(job)
        except AttributeError:
            queue.enqueue_job(job)

        # Remove job from correct registry if needed
        registry: Any
        if job.get_status() == JobStatus.DEFERRED:
            registry = DeferredJobRegistry(queue.name, queue.connection)
            registry.remove(job)
        elif job.get_status() == JobStatus.FINISHED:
            registry = FinishedJobRegistry(queue.name, queue.connection)
            registry.remove(job)
        elif job.get_status() == JobStatus.SCHEDULED:
            registry = ScheduledJobRegistry(queue.name, queue.connection)
            registry.remove(job)

        messages.info(request, f'You have successfully enqueued {job.id}')
        return redirect(rq_viewname(request, "job_detail"), queue_index, job_id)

    context_data = {
        **each_context(request),
        'queue_index': queue_index,
        'job': job,
        'queue': queue,
    }
    return render(request, 'django_rq/delete_job.html', context_data)


@never_cache
@staff_member_required
@require_POST
def stop_job(request: HttpRequest, queue_index: int, job_id: str) -> HttpResponse:
    """Stop started job"""
    queue = get_queue_by_index(queue_index)
    stopped, _ = stop_jobs(queue, job_id)
    if len(stopped) == 1:
        messages.info(request, f'You have successfully stopped {job_id}')
        return redirect(rq_viewname(request, "job_detail"), queue_index, job_id)
    else:
        messages.error(request, f'Failed to stop {job_id}')
        return redirect(rq_viewname(request, "job_detail"), queue_index, job_id)


@never_cache
@staff_member_required
def scheduler_jobs(request: HttpRequest, scheduler_index: int) -> HttpResponse:
    scheduler = get_scheduler_by_index(scheduler_index)

    items_per_page = 100
    num_jobs = scheduler.count()
    page = int(request.GET.get('page', 1))
    jobs = []

    if num_jobs > 0:
        last_page = int(ceil(num_jobs / items_per_page))
        page_range = list(range(1, last_page + 1))
        offset = items_per_page * (page - 1)
        jobs_times = scheduler.get_jobs(with_times=True, offset=offset, length=items_per_page)
        for job, time in jobs_times:
            job.next_run = time
            job.queue_index = get_queues_map().get(job.origin, 0)
            if 'cron_string' in job.meta:
                job.schedule = f"cron: '{job.meta['cron_string']}'"
            elif 'interval' in job.meta:
                job.schedule = f"interval: {job.meta['interval']}"
                if 'repeat' in job.meta:
                    job.schedule += f" repeat: {job.meta['repeat']}"
            else:
                job.schedule = 'unknown'
            jobs.append(job)
    else:
        page_range = []

    context_data = {
        **each_context(request),
        'scheduler': scheduler,
        'jobs': jobs,
        'num_jobs': num_jobs,
        'page': page,
        'page_range': page_range,
    }
    return render(request, 'django_rq/scheduler.html', context_data)



# Generated by Django 1.10.5 on 2017-02-09 15:24

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='MyModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField(unique=True)),
            ],
        ),
    ]



from secrets import compare_digest

from typing import Optional

from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404, HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render
from django.views.decorators.cache import never_cache

from . import settings as django_rq_settings
from .utils import get_cron_schedulers, get_scheduler_statistics, get_statistics
from .views import each_context

try:
    import prometheus_client

    from .contrib.prometheus import RQCollector
except ImportError:
    prometheus_client = RQCollector = None  # type: ignore[assignment, misc]

registry = None


def is_authorized(request: HttpRequest) -> bool:
    api_token = django_rq_settings.get_api_token()
    auth_header = request.headers.get("Authorization", "")
    token = None

    if auth_header.startswith("Bearer "):
        token = auth_header.removeprefix("Bearer ").strip()

    is_staff = getattr(request.user, 'is_staff', False)
    return bool(is_staff or (api_token and token and compare_digest(api_token, token)))


@never_cache
def prometheus_metrics(request):
    if not is_authorized(request):
        return JsonResponse(
            {
                "error": True,
                "description": "Missing bearer token. Set token in headers and configure RQ_API_TOKEN in settings.py",
            },
            status=401,
        )

    global registry

    if not RQCollector:  # type: ignore[truthy-function]
        raise Http404('prometheus_client has not been installed; install using extra "django-rq[prometheus]"')

    if not registry:
        registry = prometheus_client.CollectorRegistry(auto_describe=True)
        registry.register(RQCollector())

    encoder, content_type = prometheus_client.exposition.choose_encoder(request.META.get('HTTP_ACCEPT', ''))
    if 'name[]' in request.GET:
        registry = registry.restricted_registry(request.GET.getlist('name[]'))

    return HttpResponse(encoder(registry), headers={'Content-Type': content_type})


@never_cache
@staff_member_required
def stats(request: HttpRequest) -> HttpResponse:
    context_data = {
        **each_context(request),
        **get_statistics(run_maintenance_tasks=True),
        **get_scheduler_statistics(),
        "view_metrics": RQCollector is not None,
        "cron_schedulers": get_cron_schedulers(),
    }
    return render(request, 'django_rq/stats.html', context_data)


@never_cache
def stats_json(request: HttpRequest, token: Optional[str] = None) -> JsonResponse:
    if not is_authorized(request):
        api_token = django_rq_settings.get_api_token()
        if token and token == api_token:
            return JsonResponse(get_statistics())
        else:
            return JsonResponse(
                {
                    "error": True,
                    "description": "Missing bearer token. Set token in headers and configure RQ_API_TOKEN in settings.py",
                },
                status=401,
            )

    return JsonResponse(get_statistics())



import warnings
from typing import Any, Callable, Optional, Union, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connection, transaction
from redis import Redis
from rq.job import Job
from rq.queue import Queue
from rq.utils import import_attribute

from . import thread_queue
from .connection_utils import (
    filter_connection_params,
    get_connection,
    get_redis_connection,
)
from .jobs import get_job_class
from .settings import get_queues_list

VALID_COMMIT_MODES = ('auto', 'request_finished', 'on_db_commit')


def get_commit_mode() -> str:
    """
    Returns the configured commit mode.

    COMMIT_MODE is preferred and supports:
    - "on_db_commit" (default): enqueue when the database transaction commits
    - "auto": enqueue immediately
    - "request_finished": enqueue when the request finishes (legacy behavior)

    If COMMIT_MODE isn't set, fall back to AUTOCOMMIT for backwards
    compatibility (True -> "auto", False -> "request_finished").
    """
    RQ = getattr(settings, 'RQ', {})
    commit_mode = RQ.get('COMMIT_MODE')

    if not commit_mode:
        if 'AUTOCOMMIT' in RQ:
            warnings.warn('"AUTOCOMMIT" is deprecated; use "COMMIT_MODE" instead.', DeprecationWarning)
            return 'auto' if RQ.get('AUTOCOMMIT') else 'request_finished'
        return 'on_db_commit'

    if commit_mode in VALID_COMMIT_MODES:
        return commit_mode
    else:
        raise ImproperlyConfigured(f'"COMMIT_MODE" must be one of {VALID_COMMIT_MODES}.')


def get_queue_class(
    config: Optional[dict[str, Any]] = None,
    queue_class: Optional[Union[str, type[Queue]]] = None,
) -> type[Queue]:
    """
    Return queue class from config or from RQ settings, otherwise return DjangoRQ.
    If ``queue_class`` is provided, it takes priority.

    The full priority list for queue class sources:
    1. ``queue_class`` argument
    2. ``QUEUE_CLASS`` in ``config`` argument
    3. ``QUEUE_CLASS`` in base settings (``RQ``)
    """
    RQ = getattr(settings, 'RQ', {})
    if queue_class is None:
        queue_class = RQ.get('QUEUE_CLASS', DjangoRQ)
        if config:
            queue_class = config.get('QUEUE_CLASS', queue_class)

    if isinstance(queue_class, str):
        queue_class = cast(type[Queue], import_attribute(queue_class))
    return queue_class


class DjangoRQ(Queue):
    """
    A subclass of RQ's QUEUE that allows jobs to be stored temporarily to be
    enqueued later at the end of Django's request/response cycle.
    """

    def __init__(self, *args, **kwargs):
        autocommit = kwargs.pop('autocommit', None)
        commit_mode = kwargs.pop('commit_mode', None)

        if commit_mode:
            if commit_mode not in VALID_COMMIT_MODES:
                raise ImproperlyConfigured(f'commit_mode must be one of {VALID_COMMIT_MODES}.')
        elif autocommit is not None:
            warnings.warn('The "autocommit" argument is deprecated; use "commit_mode" instead.', DeprecationWarning)
            commit_mode = 'auto' if autocommit else 'request_finished'
        else:
            commit_mode = get_commit_mode()

        self._commit_mode = commit_mode

        super().__init__(*args, **kwargs)

    def original_enqueue_call(self, *args, **kwargs):
        queue_name = kwargs.get('queue_name') or self.name
        kwargs['result_ttl'] = kwargs.get('result_ttl', get_result_ttl(queue_name))

        return super().enqueue_call(*args, **kwargs)

    def enqueue_call(self, *args, **kwargs):
        if self._commit_mode == 'auto':
            return self.original_enqueue_call(*args, **kwargs)
        elif self._commit_mode == 'on_db_commit':
            if connection.in_atomic_block:
                transaction.on_commit(lambda: self.original_enqueue_call(*args, **kwargs))
            else:
                return self.original_enqueue_call(*args, **kwargs)
        else:
            thread_queue.add(self, args, kwargs)


def get_queue(
    name: str = 'default',
    default_timeout: Optional[int] = None,
    is_async: Optional[bool] = None,
    autocommit: Optional[bool] = None,
    commit_mode: Optional[str] = None,
    connection: Optional[Redis] = None,
    queue_class: Optional[Union[str, type[Queue]]] = None,
    job_class: Optional[Union[str, type[Job]]] = None,
    serializer: Any = None,
    **kwargs: Any,
) -> Queue:
    """
    Returns an rq Queue using parameters defined in ``RQ_QUEUES``
    """
    from .settings import QUEUES

    if kwargs.get('async') is not None:
        is_async = kwargs['async']
        warnings.warn('The `async` keyword is deprecated. Use `is_async` instead', DeprecationWarning)

    # If is_async is provided, use it, otherwise, get it from the configuration
    if is_async is None:
        is_async = QUEUES[name].get('ASYNC', True)
    # same for job_class
    job_class = get_job_class(job_class)

    if default_timeout is None:
        default_timeout = QUEUES[name].get('DEFAULT_TIMEOUT')
    if connection is None:
        connection = get_connection(name)
    if serializer is None:
        serializer = QUEUES[name].get('SERIALIZER')
    queue_class = get_queue_class(QUEUES[name], queue_class)
    return queue_class(
        name,
        default_timeout=default_timeout,
        connection=connection,
        is_async=is_async,
        job_class=job_class,
        autocommit=autocommit,
        commit_mode=commit_mode,
        serializer=serializer,
        **kwargs,
    )


def get_queue_by_index(index: int) -> Queue:
    """
    Returns an rq Queue using parameters defined in ``QUEUES_LIST``
    """
    config = get_queues_list()[int(index)]
    return get_queue_class(config)(
        config['name'],
        connection=get_redis_connection(config['connection_config']),
        is_async=config.get('ASYNC', True),
        serializer=config['connection_config'].get('SERIALIZER'),
    )


def get_scheduler_by_index(index: int) -> 'DjangoScheduler':
    """
    Returns an rq-scheduler Scheduler using parameters defined in ``QUEUES_LIST``
    """
    config = get_queues_list()[int(index)]
    return get_scheduler(config['name'])


def get_queues(*queue_names, **kwargs):
    """
    Return queue instances from specified queue names.
    All instances must use the same Redis connection.
    """
    from .settings import QUEUES

    if len(queue_names) <= 1:
        # Return "default" queue if no queue name is specified
        # or one queue with specified name
        return [get_queue(*queue_names, **kwargs)]

    # will return more than one queue
    # import job class only once for all queues
    kwargs['job_class'] = get_job_class(kwargs.pop('job_class', None))

    queue_params = QUEUES[queue_names[0]]
    connection_params = filter_connection_params(queue_params)
    queues = [get_queue(queue_names[0], **kwargs)]

    # do consistency checks while building return list
    for name in queue_names[1:]:
        queue = get_queue(name, **kwargs)
        if type(queue) is not type(queues[0]):
            raise ValueError(f'Queues must have the same class."{name}" and "{queue_names[0]}" have different classes')
        if connection_params != filter_connection_params(QUEUES[name]):
            raise ValueError(
                f'Queues must have the same redis connection."{name}" and "{queue_names[0]}" have different connections'
            )
        queues.append(queue)

    return queues


def enqueue(func: Callable, *args, **kwargs) -> Job:
    """
    A convenience function to put a job in the default queue. Usage::

    from django_rq import enqueue
    enqueue(func, *args, **kwargs)
    """
    return get_queue().enqueue(func, *args, **kwargs)


def get_result_ttl(name: str = 'default') -> Optional[int]:
    """
    Returns the result ttl from RQ_QUEUES if found, otherwise from RQ
    """
    from .settings import QUEUES

    RQ = getattr(settings, 'RQ', {})
    return QUEUES[name].get('DEFAULT_RESULT_TTL', RQ.get('DEFAULT_RESULT_TTL'))


"""
If rq_scheduler is installed, provide a ``get_scheduler`` function that
behaves like ``get_connection``, except that it returns a ``Scheduler``
instance instead of a ``Queue`` instance.
"""
try:
    from rq_scheduler import Scheduler

    class DjangoScheduler(Scheduler):
        """
        Use settings ``DEFAULT_RESULT_TTL`` from ``RQ``
        and ``DEFAULT_TIMEOUT`` from ``RQ_QUEUES`` if configured.
        """

        def _create_job(self, *args, **kwargs):
            from .settings import QUEUES

            if kwargs.get('timeout') is None:
                queue_name = kwargs.get('queue_name') or self.queue_name
                kwargs['timeout'] = QUEUES[queue_name].get('DEFAULT_TIMEOUT')

            if kwargs.get('result_ttl') is None:
                kwargs['result_ttl'] = getattr(settings, 'RQ', {}).get('DEFAULT_RESULT_TTL')

            return super()._create_job(*args, **kwargs)

    def get_scheduler(
        name: str = 'default',
        queue: Optional[Queue] = None,
        interval: int = 60,
        connection: Optional[Redis] = None,
    ) -> DjangoScheduler:
        """
        Returns an RQ Scheduler instance using parameters defined in
        ``RQ_QUEUES``
        """
        RQ = getattr(settings, 'RQ', {})
        scheduler_class = RQ.get('SCHEDULER_CLASS', DjangoScheduler)

        if isinstance(scheduler_class, str):
            scheduler_class = import_attribute(scheduler_class)

        if connection is None:
            connection = get_connection(name)

        if queue is None:
            queue = get_queue(name, connection=connection)

        return scheduler_class(
            queue_name=name, interval=interval, queue=queue, job_class=queue.job_class, connection=connection
        )

except ImportError:

    def get_scheduler(*args, **kwargs):  # type: ignore[misc]
        raise ImproperlyConfigured('rq_scheduler not installed')



from typing import Optional, Union, cast

from django.conf import settings
from rq.job import Job
from rq.utils import import_attribute


def get_job_class(job_class: Optional[Union[str, type[Job]]] = None) -> type[Job]:
    """
    Return job class from RQ settings, otherwise return Job.
    If `job_class` is not None, it is used as an override (can be
    python import path as string).
    """
    RQ = getattr(settings, 'RQ', {})

    if not job_class:
        job_class = RQ.get('JOB_CLASS')

    # Ensure we never have None (in case someone explicitly sets JOB_CLASS to None)
    if not job_class:
        job_class = Job

    if isinstance(job_class, str):
        job_class = cast(type[Job], import_attribute(job_class))
    return job_class



from rq.job import JobStatus

from ..connection_utils import filter_connection_params, get_connection, get_unique_connection_configs
from ..queues import get_queue
from ..workers import get_worker_class

try:
    from prometheus_client import Summary
    from prometheus_client.core import CounterMetricFamily, GaugeMetricFamily

    class RQCollector:
        """RQ stats collector"""

        summary = Summary('rq_request_processing_seconds_total', 'Time spent collecting RQ data')

        def collect(self):
            from ..settings import QUEUES

            with self.summary.time():
                rq_workers = GaugeMetricFamily('rq_workers', 'RQ workers', labels=['name', 'state', 'queues'])
                rq_job_successful_total = CounterMetricFamily(
                    'rq_job_successful_total', 'RQ successful job count', labels=['name', 'queues']
                )
                rq_job_failed_total = CounterMetricFamily(
                    'rq_job_failed_total', 'RQ failed job count', labels=['name', 'queues']
                )
                rq_working_seconds_total = CounterMetricFamily(
                    'rq_working_seconds_total', 'RQ total working time', labels=['name', 'queues']
                )

                rq_jobs = GaugeMetricFamily('rq_jobs', 'RQ jobs by status', labels=['queue', 'status'])

                worker_class = get_worker_class()
                unique_configs = get_unique_connection_configs()
                connections = {}
                for queue_name, config in QUEUES.items():
                    index = unique_configs.index(filter_connection_params(config))
                    if index not in connections:
                        connections[index] = connection = get_connection(queue_name)

                        for worker in worker_class.all(connection):
                            name = worker.name
                            label_queues = ','.join(worker.queue_names())
                            rq_workers.add_metric([name, worker.get_state(), label_queues], 1)
                            rq_job_successful_total.add_metric([name, label_queues], worker.successful_job_count)
                            rq_job_failed_total.add_metric([name, label_queues], worker.failed_job_count)
                            rq_working_seconds_total.add_metric([name, label_queues], worker.total_working_time)
                    else:
                        connection = connections[index]

                    queue = get_queue(queue_name, connection=connection)
                    rq_jobs.add_metric([queue_name, JobStatus.QUEUED], queue.count)
                    rq_jobs.add_metric([queue_name, JobStatus.STARTED], queue.started_job_registry.count)
                    rq_jobs.add_metric([queue_name, JobStatus.FINISHED], queue.finished_job_registry.count)
                    rq_jobs.add_metric([queue_name, JobStatus.FAILED], queue.failed_job_registry.count)
                    rq_jobs.add_metric([queue_name, JobStatus.DEFERRED], queue.deferred_job_registry.count)
                    rq_jobs.add_metric([queue_name, JobStatus.SCHEDULED], queue.scheduled_job_registry.count)

                yield rq_workers
                yield rq_job_successful_total
                yield rq_job_failed_total
                yield rq_working_seconds_total
                yield rq_jobs

except ImportError:
    RQCollector = None  # type: ignore[assignment, misc]



from django.core.signals import got_request_exception, request_finished
from django.db import models

from . import thread_queue
from .queues import get_commit_mode

# Wire up request finished/exception signal only for the request_finished mode
if get_commit_mode() == 'request_finished':
    request_finished.connect(thread_queue.commit)
    got_request_exception.connect(thread_queue.clear)


class Dashboard(models.Model):
    """
    Admin-only model for Django-RQ dashboard integration.
    """

    class Meta:
        managed = False  # No database table - admin integration only
        default_permissions = ()
        permissions = [['view', 'Access admin page']]
        verbose_name = 'Django-RQ'
        verbose_name_plural = 'Django-RQ'



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass

from redis import Redis
from redis.exceptions import RedisError


@dataclass(frozen=True)
class RedisConfig:
    host: str
    port: int
    db: int


def _find_empty_databases(host: str, port: int, *, required: int) -> Sequence[int]:
    """
    Find a set of empty Redis databases that can be used for tests.

    Raises RuntimeError if we cannot connect or if we cannot find the requested
    number of empty databases. Sentinel ports are intentionally not probed.
    """

    empty_databases: list[int] = []

    for db in range(16):
        try:
            if Redis(host=host, port=port, db=db).dbsize() == 0:
                empty_databases.append(db)
                if len(empty_databases) == required:
                    break
        except RedisError as exc:  # pragma: no cover - defensive guard
            raise RuntimeError(
                f"Refusing to run tests: unable to inspect Redis database {db} on {host}:{port}"
            ) from exc

    if len(empty_databases) < required:
        raise RuntimeError(
            f"Refusing to run tests: need {required} empty Redis databases on {host}:{port}, "
            f"but only found {len(empty_databases)}."
        )

    return empty_databases


REDIS_HOST = 'localhost'
REDIS_PORT = 6379
databases = _find_empty_databases(REDIS_HOST, REDIS_PORT, required=3)

REDIS_CONFIG_1 = RedisConfig(
    host=REDIS_HOST,
    port=REDIS_PORT,
    db=databases[0],
)
REDIS_CONFIG_2 = RedisConfig(
    host=REDIS_HOST,
    port=REDIS_PORT,
    db=databases[1],
)
REDIS_CONFIG_3 = RedisConfig(
    host=REDIS_HOST,
    port=REDIS_PORT,
    db=databases[2],
)
